/*
 * Windows Shell Item functions
 *
 * Copyright (c) 2010-2013, Joachim Metz <joachim.metz@gmail.com>
 *
 * Refer to AUTHORS for acknowledgements.
 *
 * This software is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <common.h>
#include <byte_stream.h>
#include <memory.h>
#include <types.h>

#include "libfwsi_definitions.h"
#include "libfwsi_file_attributes.h"
#include "libfwsi_folder_identifier.h"
#include "libfwsi_item.h"
#include "libfwsi_libcerror.h"
#include "libfwsi_libcnotify.h"
#include "libfwsi_libcstring.h"
#include "libfwsi_libfdatetime.h"
#include "libfwsi_libfguid.h"
#include "libfwsi_libuna.h"
#include "libfwsi_types.h"

/* Creates an item
 * Make sure the value item is referencing, is set to NULL
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_initialize(
     libfwsi_item_t **item,
     libcerror_error_t **error )
{
	libfwsi_internal_item_t *internal_item = NULL;
	static char *function                  = "libfwsi_item_initialize";

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	if( *item != NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_RUNTIME,
		 LIBCERROR_RUNTIME_ERROR_VALUE_ALREADY_SET,
		 "%s: invalid item value already set.",
		 function );

		return( -1 );
	}
	internal_item = memory_allocate_structure(
	                 libfwsi_internal_item_t );

	if( internal_item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
		 "%s: unable to create item.",
		 function );

		goto on_error;
	}
	if( memory_set(
	     internal_item,
	     0,
	     sizeof( libfwsi_internal_item_t ) ) == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_MEMORY,
		 LIBCERROR_MEMORY_ERROR_SET_FAILED,
		 "%s: unable to clear item.",
		 function );

		goto on_error;
	}
	*item = (libfwsi_item_t *) internal_item;

	return( 1 );

on_error:
	if( internal_item != NULL )
	{
		memory_free(
		 internal_item );
	}
	return( -1 );
}

/* Frees an item
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_free(
     libfwsi_item_t **item,
     libcerror_error_t **error )
{
	libfwsi_internal_item_t *internal_item = NULL;
	static char *function                  = "libfwsi_item_free";

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	if( *item != NULL )
	{
		internal_item = (libfwsi_internal_item_t *) *item;
		*item         = NULL;

		memory_free(
		 internal_item );
	}
	return( 1 );
}

/* Copies a shell item from a byte stream
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_copy_from_byte_stream(
     libfwsi_item_t *item,
     const uint8_t *byte_stream,
     size_t byte_stream_size,
     libcerror_error_t **error )
{
	libfwsi_internal_item_t *internal_item      = NULL;
	static char *function                       = "libfwsi_item_copy_from_byte_stream";
	size_t string_alignment_size                = 0;
	size_t string_size                          = 0;
	size_t shell_item_data_size                 = 0;
	uint16_t data_size                          = 0;
	uint16_t extension_size                     = 0;
	uint16_t extension_version                  = 0;
	uint16_t identifier_size                    = 0;
	uint16_t long_string_size                   = 0;

#if defined( HAVE_DEBUG_OUTPUT )
	libcstring_system_character_t fat_date_time_string[ 32 ];
        libcstring_system_character_t guid_string[ 48 ];

	libcstring_system_character_t *value_string = NULL;
	libfdatetime_fat_date_time_t *fat_date_time = NULL;
        libfguid_identifier_t *guid                 = NULL;
	size_t value_string_size                    = 0;
	uint32_t value_32bit                        = 0;
	uint16_t value_16bit                        = 0;
	int result                                  = 0;

	/* TODO */
	int ascii_codepage                          = LIBUNA_CODEPAGE_WINDOWS_1252;
#endif

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	internal_item = (libfwsi_internal_item_t *) item;

	if( byte_stream == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid byte stream.",
		 function );

		return( -1 );
	}
	if( byte_stream_size < 2 )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: byte stream too small.",
		 function );

		return( -1 );
	}
	if( byte_stream_size > (size_t) SSIZE_MAX )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_EXCEEDS_MAXIMUM,
		 "%s: byte stream size exceeds maximum.",
		 function );

		return( -1 );
	}
	byte_stream_copy_to_uint16_little_endian(
	 byte_stream,
	 internal_item->size );

#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "%s: shell item size\t\t\t: %" PRIu16 "\n",
		 function,
		 internal_item->size );
	}
#endif
	byte_stream      += 2;
	byte_stream_size -= 2;

	if( byte_stream_size < (size_t) internal_item->size )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_VALUE_TOO_SMALL,
		 "%s: byte stream too small.",
		 function );

		return( -1 );
	}
	if( internal_item->size > 2 )
	{
#if defined( HAVE_DEBUG_OUTPUT )
		if( libcnotify_verbose != 0 )
		{
			libcnotify_printf(
			 "%s: shell item data:\n",
			 function );
			libcnotify_print_data(
			 byte_stream,
			 (size_t) ( internal_item->size - 2 ),
			 0 );
		}
#endif
		internal_item->type = *byte_stream;

#if defined( HAVE_DEBUG_OUTPUT )
		if( libcnotify_verbose != 0 )
		{
			libcnotify_printf(
			 "%s: shell item type\t\t\t: 0x%02" PRIx8 "\n",
			 function,
			 internal_item->type );

/* TODO add type debug string */
		}
#endif
		byte_stream      += 1;
		byte_stream_size -= 1;

		switch( internal_item->type )
		{
			case LIBFWSI_ITEM_TYPE_UNKNOWN_0x00:
				shell_item_data_size = internal_item->size - 3;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: shell item flags\t\t\t: 0x%02" PRIx8 "\n",
					 function,
					 *byte_stream );
				}
#endif
				byte_stream          += 1;
				byte_stream_size     -= 1;
				shell_item_data_size -= 1;

				byte_stream_copy_to_uint16_little_endian(
				 byte_stream,
				 data_size );

				if( data_size < ( internal_item->size - 4 ) )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: shell item data size\t\t: %" PRIu16 "\n",
						 function,
						 data_size );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 byte_stream,
						 value_32bit );

						libcnotify_printf(
						 "%s: shell item unknown1\t\t\t: 0x%08" PRIx32 "\n",
						 function,
						 value_32bit );
					}
#endif
					byte_stream          += 4;
					byte_stream_size     -= 4;
					shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: shell item unknown1 size\t\t: %" PRIu16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 identifier_size );

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: shell item identifier size\t\t: %" PRIu16 "\n",
						 function,
						 identifier_size );
					}
#endif
					byte_stream          += 2;
					byte_stream_size     -= 2;
					shell_item_data_size -= 2;

					if( identifier_size == 16 )
					{
#if defined( HAVE_DEBUG_OUTPUT )
						if( libcnotify_verbose != 0 )
						{
							if( libfguid_identifier_initialize(
							     &guid,
							     error ) != 1 )
							{
								libcerror_error_set(
								 error,
								 LIBCERROR_ERROR_DOMAIN_RUNTIME,
								 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
								 "%s: unable to create GUID.",
								 function );

								return( -1 );
							}
							if( libfguid_identifier_copy_from_byte_stream(
							     guid,
							     byte_stream,
							     byte_stream_size,
							     LIBFGUID_ENDIAN_LITTLE,
							     error ) != 1 )
							{
								libcerror_error_set(
								 error,
								 LIBCERROR_ERROR_DOMAIN_RUNTIME,
								 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
								 "%s: unable to copy byte stream to GUID.",
								 function );

								libfguid_identifier_free(
								 &guid,
								 NULL );

								return( -1 );
							}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
							result = libfguid_identifier_copy_to_utf16_string(
								  guid,
								  (uint16_t *) guid_string,
								  48,
								  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
								  error );
#else
							result = libfguid_identifier_copy_to_utf8_string(
								  guid,
								  (uint8_t *) guid_string,
								  48,
								  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
								  error );
#endif
							if( result != 1 )
							{
								libcerror_error_set(
								 error,
								 LIBCERROR_ERROR_DOMAIN_RUNTIME,
								 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
								 "%s: unable to copy GUID to string.",
								 function );

								libfguid_identifier_free(
								 &guid,
								 NULL );

								return( -1 );
							}
							if( libfguid_identifier_free(
							     &guid,
							     error ) != 1 )
							{
								libcerror_error_set(
								 error,
								 LIBCERROR_ERROR_DOMAIN_RUNTIME,
								 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
								 "%s: unable to free GUID.",
								 function );

								return( -1 );
							}
							libcnotify_printf(
							 "%s: shell item folder identifier\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
							 function,
							 guid_string );
							libcnotify_printf(
							 "%s: shell item folder name\t\t: %s\n",
							 function,
							 libfwsi_folder_identifier_get_name(
							  byte_stream ) );
						}
#endif
						byte_stream          += 16;
						byte_stream_size     -= 16;
						shell_item_data_size -= 16;
					}
				}
				break;

			case 0x30:
			case LIBFWSI_ITEM_TYPE_FILE_ENTRY_DIRECTORY:
			case LIBFWSI_ITEM_TYPE_FILE_ENTRY_FILE:
				shell_item_data_size = internal_item->size - 3;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					if( libfdatetime_fat_date_time_initialize(
					     &fat_date_time,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
						 "%s: unable to create FAT date time.",
						 function );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item flags\t\t\t: 0x%02" PRIx8 "\n",
					 function,
					 *byte_stream );
				}
#endif
				byte_stream          += 1;
				byte_stream_size     -= 1;
				shell_item_data_size -= 1;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: shell item file size\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					if( libfdatetime_fat_date_time_copy_from_byte_stream(
					     fat_date_time,
					     byte_stream,
					     byte_stream_size,
					     LIBFDATETIME_ENDIAN_LITTLE,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to FAT date time.",
						 function );

						libfdatetime_fat_date_time_free(
						 &fat_date_time,
						 NULL );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libfdatetime_fat_date_time_copy_to_utf16_string(
						  fat_date_time,
						  (uint16_t *) fat_date_time_string,
						  32,
						  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
						  error );
#else
					result = libfdatetime_fat_date_time_copy_to_utf8_string(
						  fat_date_time,
						  (uint8_t *) fat_date_time_string,
						  32,
						  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy FAT date time to string.",
						 function );

						libfdatetime_fat_date_time_free(
						 &fat_date_time,
						 NULL );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item modification time\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
					 function,
					 fat_date_time_string );
				}
#endif
				byte_stream          += 4;
				byte_stream_size     -= 4;
				shell_item_data_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: shell item file attribute flags\t: 0x%04" PRIx16 "\n",
					 function,
					 value_16bit );
					libfwsi_file_attributes_print(
					 value_16bit );
					libcnotify_printf(
					 "\n" );
				}
#endif
				byte_stream          += 2;
				byte_stream_size     -= 2;
				shell_item_data_size -= 2;

/* TODO determine win2k or win2k3 */

				/* Determine the size of the primary name (Windows 2000)
				 * or the short name (Windows 2003 or later)
				 */
				for( string_size = 0;
				     string_size < byte_stream_size;
				     string_size++ )
				{
					if( byte_stream[ string_size ] == 0 )
					{
						break;
					}
				}
				string_size++;

				string_alignment_size = string_size % 2;

				if( ( shell_item_data_size - ( string_size + string_alignment_size ) ) >= 2 )
				{
					/* Look ahead if the extension size makes sense
					 * if not we're dealing with a Windows 2000 entry
					 * otherwise a Windows 2003 or later
					 */
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream + string_size + string_alignment_size,
					 extension_size );
				}
				if( ( ( shell_item_data_size - ( string_size + string_alignment_size ) ) < 2 )
				 || ( extension_size > shell_item_data_size ) )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_byte_stream(
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_byte_stream(
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of primary name string.",
							 function );

							return( -1 );
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid primary name string size value exceeds maximum.",
							 function );

							return( -1 );
						}
						value_string = libcstring_system_string_allocate(
								value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create primary name string.",
							 function );

							return( -1 );
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_byte_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  error );
#else
						result = libuna_utf8_string_copy_from_byte_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set primary name string.",
							 function );

							memory_free(
							 value_string );

							return( -1 );
						}
						libcnotify_printf(
						 "%s: shell item primary name\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );
					}
#endif
					byte_stream          += string_size;
					byte_stream_size     -= string_size;
					shell_item_data_size -= string_size;

					/* Determine the size of the secondary name
					 */
					for( string_size = 0;
					     string_size < byte_stream_size;
					     string_size++ )
					{
						if( byte_stream[ string_size ] == 0 )
						{
							break;
						}
					}
					string_size++;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_byte_stream(
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_byte_stream(
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of secondary name string.",
							 function );

							return( -1 );
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid secondary name string size value exceeds maximum.",
							 function );

							return( -1 );
						}
						value_string = libcstring_system_string_allocate(
								value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create secondary name string.",
							 function );

							return( -1 );
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_byte_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  error );
#else
						result = libuna_utf8_string_copy_from_byte_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set secondary name string.",
							 function );

							memory_free(
							 value_string );

							return( -1 );
						}
						libcnotify_printf(
						 "%s: shell item secondary name\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );
					}
#endif
					byte_stream          += string_size;
					byte_stream_size     -= string_size;
					shell_item_data_size -= string_size;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						if( shell_item_data_size > 0 )
						{
							libcnotify_printf(
							 "%s: shell item trailing data:\n",
							 function );
							libcnotify_print_data(
							 byte_stream,
							 shell_item_data_size,
							 0 );
						}
					}
#endif
					break;
				}
				else
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_byte_stream(
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_byte_stream(
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of short name string.",
							 function );

							return( -1 );
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid short name string size value exceeds maximum.",
							 function );

							return( -1 );
						}
						value_string = libcstring_system_string_allocate(
								value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create short name string.",
							 function );

							return( -1 );
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_byte_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  error );
#else
						result = libuna_utf8_string_copy_from_byte_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  ascii_codepage,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set short name string.",
							 function );

							memory_free(
							 value_string );

							return( -1 );
						}
						libcnotify_printf(
						 "%s: shell item short name\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );
					}
#endif
					byte_stream          += string_size + string_alignment_size;
					byte_stream_size     -= string_size + string_alignment_size;
					shell_item_data_size -= string_size + string_alignment_size;
				}
				byte_stream_copy_to_uint16_little_endian(
				 byte_stream,
				 extension_size );

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: shell item extension size\t\t: %" PRIu16 "\n",
					 function,
					 extension_size );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

				byte_stream_copy_to_uint16_little_endian(
				 byte_stream,
				 extension_version );

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: shell item extension version\t: %" PRIu16 "\n",
					 function,
					 extension_version );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: shell item unknown1\t\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: shell item unknown2\t\t\t: 0x%04" PRIx16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					if( libfdatetime_fat_date_time_copy_from_byte_stream(
					     fat_date_time,
					     byte_stream,
					     byte_stream_size,
					     LIBFDATETIME_ENDIAN_LITTLE,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to FAT date time.",
						 function );

						libfdatetime_fat_date_time_free(
						 &fat_date_time,
						 NULL );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libfdatetime_fat_date_time_copy_to_utf16_string(
						  fat_date_time,
						  (uint16_t *) fat_date_time_string,
						  32,
						  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
						  error );
#else
					result = libfdatetime_fat_date_time_copy_to_utf8_string(
						  fat_date_time,
						  (uint8_t *) fat_date_time_string,
						  32,
						  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy FAT date time to string.",
						 function );

						libfdatetime_fat_date_time_free(
						 &fat_date_time,
						 NULL );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item creation time\t\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
					 function,
					 fat_date_time_string );
				}
#endif
				byte_stream      += 4;
				byte_stream_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					if( libfdatetime_fat_date_time_copy_from_byte_stream(
					     fat_date_time,
					     byte_stream,
					     byte_stream_size,
					     LIBFDATETIME_ENDIAN_LITTLE,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to FAT date time.",
						 function );

						libfdatetime_fat_date_time_free(
						 &fat_date_time,
						 NULL );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libfdatetime_fat_date_time_copy_to_utf16_string(
						  fat_date_time,
						  (uint16_t *) fat_date_time_string,
						  32,
						  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
						  error );
#else
					result = libfdatetime_fat_date_time_copy_to_utf8_string(
						  fat_date_time,
						  (uint8_t *) fat_date_time_string,
						  32,
						  LIBFDATETIME_STRING_FORMAT_TYPE_CTIME | LIBFDATETIME_STRING_FORMAT_FLAG_DATE_TIME,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy FAT date time to string.",
						 function );

						libfdatetime_fat_date_time_free(
						 &fat_date_time,
						 NULL );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item access time\t\t: %" PRIs_LIBCSTRING_SYSTEM " UTC\n",
					 function,
					 fat_date_time_string );
				}
#endif
				byte_stream      += 4;
				byte_stream_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: shell item unknown3\t\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream      += 4;
				byte_stream_size -= 4;

				if( extension_version >= 7 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 byte_stream,
						 value_32bit );

						libcnotify_printf(
						 "%s: shell item MFT entry index\t\t: %" PRIu32 "\n",
						 function,
						 value_32bit );
					}
#endif
					byte_stream      += 4;
					byte_stream_size -= 4;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: shell item unknown4\t\t\t: %" PRIu16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream      += 2;
					byte_stream_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint16_little_endian(
						 byte_stream,
						 value_16bit );

						libcnotify_printf(
						 "%s: shell item MFT sequence\t\t: %" PRIu16 "\n",
						 function,
						 value_16bit );
					}
#endif
					byte_stream      += 2;
					byte_stream_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: shell item unknown5:\n",
						 function );
						libcnotify_print_data(
						 byte_stream,
						 8,
						 0 );
					}
#endif
					byte_stream      += 8;
					byte_stream_size -= 8;

					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 long_string_size );

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: shell item long string size\t\t: %" PRIu16 "\n",
						 function,
						 long_string_size );
					}
#endif
					byte_stream      += 2;
					byte_stream_size -= 2;
				}
				if( extension_version >= 8 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						libcnotify_printf(
						 "%s: shell item unknown6:\n",
						 function );
						libcnotify_print_data(
						 byte_stream,
						 4,
						 0 );
					}
#endif
					byte_stream      += 4;
					byte_stream_size -= 4;
				}
				/* Determine the long name size
				 */
				for( string_size = 0;
				     ( string_size + 1 ) < byte_stream_size;
				     string_size += 2 )
				{
					if( ( byte_stream[ string_size ] == 0 )
					 && ( byte_stream[ string_size + 1 ] == 0 ) )
					{
						break;
					}
				}
				string_size += 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_utf16_stream(
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_utf16_stream(
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  &value_string_size,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine size of long name string.",
						 function );

						return( -1 );
					}
					if( ( value_string_size > (size_t) SSIZE_MAX )
					 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
						 "%s: invalid long name string size value exceeds maximum.",
						 function );

						return( -1 );
					}
					value_string = libcstring_system_string_allocate(
					                value_string_size );

					if( value_string == NULL )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_MEMORY,
						 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
						 "%s: unable to create long name string.",
						 function );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_utf16_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  error );
#else
					result = libuna_utf8_string_copy_from_utf16_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
						 "%s: unable to set long name string.",
						 function );

						memory_free(
						 value_string );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item long name\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
					 function,
					 value_string );

					memory_free(
					 value_string );
				}
#endif
				byte_stream      += string_size;
				byte_stream_size -= string_size;

				if( long_string_size > 0 )
				{
					/* Determine the additional long name size
					 */
					for( string_size = 0;
					     ( string_size + 1 ) < byte_stream_size;
					     string_size += 2 )
					{
						if( ( byte_stream[ string_size ] == 0 )
						 && ( byte_stream[ string_size + 1 ] == 0 ) )
						{
							break;
						}
					}
					string_size += 2;

#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_size_from_utf16_stream(
							  byte_stream,
							  byte_stream_size,
							  LIBUNA_ENDIAN_LITTLE,
							  &value_string_size,
							  error );
#else
						result = libuna_utf8_string_size_from_utf16_stream(
							  byte_stream,
							  byte_stream_size,
							  LIBUNA_ENDIAN_LITTLE,
							  &value_string_size,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
							 "%s: unable to determine size of additional long name string.",
							 function );

							return( -1 );
						}
						if( ( value_string_size > (size_t) SSIZE_MAX )
						 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
							 "%s: invalid additional long name string size value exceeds maximum.",
							 function );

							return( -1 );
						}
						value_string = libcstring_system_string_allocate(
						                value_string_size );

						if( value_string == NULL )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_MEMORY,
							 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
							 "%s: unable to create additional long name string.",
							 function );

							return( -1 );
						}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
						result = libuna_utf16_string_copy_from_utf16_stream(
							  (libuna_utf16_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  LIBUNA_ENDIAN_LITTLE,
							  error );
#else
						result = libuna_utf8_string_copy_from_utf16_stream(
							  (libuna_utf8_character_t *) value_string,
							  value_string_size,
							  byte_stream,
							  byte_stream_size,
							  LIBUNA_ENDIAN_LITTLE,
							  error );
#endif
						if( result != 1 )
						{
							libcerror_error_set(
							 error,
							 LIBCERROR_ERROR_DOMAIN_RUNTIME,
							 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
							 "%s: unable to set additional long name string.",
							 function );

							memory_free(
							 value_string );

							return( -1 );
						}
						libcnotify_printf(
						 "%s: shell item additional long name\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
						 function,
						 value_string );

						memory_free(
						 value_string );
					}
#endif
					byte_stream      += string_size;
					byte_stream_size -= string_size;
				}

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: shell item unknown7\t\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					if( libfdatetime_fat_date_time_free(
					     &fat_date_time,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
						 "%s: unable to free FAT date time.",
						 function );

						return( -1 );
					}
				}
#endif
				break;

			case LIBFWSI_ITEM_TYPE_FOLDER_IDENTIFIER:
				if( internal_item->size != 20 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
					 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
					 "%s: unsupported folder identifier size.",
					 function );

					return( -1 );
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: shell item flags\t\t\t: 0x%02" PRIx8 "\n",
					 function,
					 *byte_stream );
				}
#endif
				byte_stream      += 1;
				byte_stream_size -= 1;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					if( libfguid_identifier_initialize(
					     &guid,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_INITIALIZE_FAILED,
						 "%s: unable to create GUID.",
						 function );

						return( -1 );
					}
					if( libfguid_identifier_copy_from_byte_stream(
					     guid,
					     byte_stream,
					     byte_stream_size,
					     LIBFGUID_ENDIAN_LITTLE,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy byte stream to GUID.",
						 function );

						libfguid_identifier_free(
						 &guid,
						 NULL );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libfguid_identifier_copy_to_utf16_string(
						  guid,
						  (uint16_t *) guid_string,
						  48,
						  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
						  error );
#else
					result = libfguid_identifier_copy_to_utf8_string(
						  guid,
						  (uint8_t *) guid_string,
						  48,
						  LIBFGUID_STRING_FORMAT_FLAG_USE_UPPER_CASE | LIBFGUID_STRING_FORMAT_FLAG_USE_SURROUNDING_BRACES,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_COPY_FAILED,
						 "%s: unable to copy GUID to string.",
						 function );

						libfguid_identifier_free(
						 &guid,
						 NULL );

						return( -1 );
					}
					if( libfguid_identifier_free(
					     &guid,
					     error ) != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_FINALIZE_FAILED,
						 "%s: unable to free GUID.",
						 function );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item folder identifier\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
					 function,
					 guid_string );
					libcnotify_printf(
					 "%s: shell item folder name\t\t: %s\n",
					 function,
					 libfwsi_folder_identifier_get_name(
					  byte_stream ) );
				}
#endif
				break;

			case 0x23:
			case 0x25:
			case 0x29:
			case LIBFWSI_ITEM_TYPE_VOLUME_NAME:
				if( internal_item->size != 25 )
				{
					libcerror_error_set(
					 error,
					 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
					 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
					 "%s: unsupported volume name size.",
					 function );

					return( -1 );
				}
#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_byte_stream(
						  byte_stream,
						  20,
						  ascii_codepage,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_byte_stream(
						  byte_stream,
						  20,
						  ascii_codepage,
						  &value_string_size,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine size of volume name string.",
						 function );

						return( -1 );
					}
					if( ( value_string_size > (size_t) SSIZE_MAX )
					 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
						 "%s: invalid volume name string size value exceeds maximum.",
						 function );

						return( -1 );
					}
					value_string = libcstring_system_string_allocate(
					                value_string_size );

					if( value_string == NULL )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_MEMORY,
						 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
						 "%s: unable to create volume name string.",
						 function );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_byte_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  20,
						  ascii_codepage,
						  error );
#else
					result = libuna_utf8_string_copy_from_byte_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  20,
						  ascii_codepage,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
						 "%s: unable to set volume name string.",
						 function );

						memory_free(
						 value_string );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item volume name\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
					 function,
					 value_string );

					memory_free(
					 value_string );
				}
#endif
				byte_stream      += 20;
				byte_stream_size -= 20;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: shell item unknown1\t\t\t: 0x%04" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

				break;

			case LIBFWSI_ITEM_TYPE_URI:
				shell_item_data_size = internal_item->size - 3;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					libcnotify_printf(
					 "%s: shell item flags\t\t\t: 0x%02" PRIx8 "\n",
					 function,
					 *byte_stream );
				}
#endif
				byte_stream      += 1;
				byte_stream_size -= 1;

				shell_item_data_size -= 1;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint32_little_endian(
					 byte_stream,
					 value_32bit );

					libcnotify_printf(
					 "%s: shell item unknown1\t\t\t: %" PRIu32 "\n",
					 function,
					 value_32bit );
				}
#endif
				byte_stream      += 4;
				byte_stream_size -= 4;

				shell_item_data_size -= 4;

				/* Determine the URI size
				 */
				for( string_size = 0;
				     ( string_size + 1 ) < byte_stream_size;
				     string_size += 2 )
				{
					if( ( byte_stream[ string_size ] == 0 )
					 && ( byte_stream[ string_size + 1 ] == 0 ) )
					{
						break;
					}
				}
				string_size += 2;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_size_from_utf16_stream(
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  &value_string_size,
						  error );
#else
					result = libuna_utf8_string_size_from_utf16_stream(
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  &value_string_size,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_GET_FAILED,
						 "%s: unable to determine size of URI string.",
						 function );

						return( -1 );
					}
					if( ( value_string_size > (size_t) SSIZE_MAX )
					 || ( ( sizeof( libcstring_system_character_t ) * value_string_size )  > (size_t) SSIZE_MAX ) )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_VALUE_EXCEEDS_MAXIMUM,
						 "%s: invalid URI string size value exceeds maximum.",
						 function );

						return( -1 );
					}
					value_string = libcstring_system_string_allocate(
					                value_string_size );

					if( value_string == NULL )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_MEMORY,
						 LIBCERROR_MEMORY_ERROR_INSUFFICIENT,
						 "%s: unable to create URI string.",
						 function );

						return( -1 );
					}
#if defined( LIBCSTRING_HAVE_WIDE_SYSTEM_CHARACTER )
					result = libuna_utf16_string_copy_from_utf16_stream(
						  (libuna_utf16_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  error );
#else
					result = libuna_utf8_string_copy_from_utf16_stream(
						  (libuna_utf8_character_t *) value_string,
						  value_string_size,
						  byte_stream,
						  byte_stream_size,
						  LIBUNA_ENDIAN_LITTLE,
						  error );
#endif
					if( result != 1 )
					{
						libcerror_error_set(
						 error,
						 LIBCERROR_ERROR_DOMAIN_RUNTIME,
						 LIBCERROR_RUNTIME_ERROR_SET_FAILED,
						 "%s: unable to set URI string.",
						 function );

						memory_free(
						 value_string );

						return( -1 );
					}
					libcnotify_printf(
					 "%s: shell item URI\t\t\t: %" PRIs_LIBCSTRING_SYSTEM "\n",
					 function,
					 value_string );

					memory_free(
					 value_string );
				}
#endif
				byte_stream      += string_size;
				byte_stream_size -= string_size;

				shell_item_data_size -= string_size;

#if defined( HAVE_DEBUG_OUTPUT )
				if( libcnotify_verbose != 0 )
				{
					byte_stream_copy_to_uint16_little_endian(
					 byte_stream,
					 value_16bit );

					libcnotify_printf(
					 "%s: shell item unknown2\t\t\t: %" PRIu16 "\n",
					 function,
					 value_16bit );
				}
#endif
				byte_stream      += 2;
				byte_stream_size -= 2;

				shell_item_data_size -= 2;

				if( shell_item_data_size > 0 )
				{
#if defined( HAVE_DEBUG_OUTPUT )
					if( libcnotify_verbose != 0 )
					{
						byte_stream_copy_to_uint32_little_endian(
						 byte_stream,
						 value_32bit );

						libcnotify_printf(
						 "%s: shell item unknown3\t\t\t: %" PRIu32 "\n",
						 function,
						 value_32bit );
					}
#endif
					byte_stream      += 4;
					byte_stream_size -= 4;
				}
				break;

			default:
/* TODO
				libcerror_error_set(
				 error,
				 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
				 LIBCERROR_ARGUMENT_ERROR_UNSUPPORTED_VALUE,
				 "%s: unsupported shell item type: 0x%02" PRIx8 ".",
				 function,
				 internal_item->type );

				return( -1 );
*/
				break;
		}
	}
#if defined( HAVE_DEBUG_OUTPUT )
	if( libcnotify_verbose != 0 )
	{
		libcnotify_printf(
		 "\n" );
	}
#endif
	return( 1 );
}

/* Retrieved the item typ
 * Returns 1 if successful or -1 on error
 */
int libfwsi_item_get_type(
     libfwsi_item_t *item,
     uint8_t *item_type,
     libcerror_error_t **error )
{
	libfwsi_internal_item_t *internal_item = NULL;
	static char *function                  = "libfwsi_item_get_type";

	if( item == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item.",
		 function );

		return( -1 );
	}
	internal_item = (libfwsi_internal_item_t *) item;

	if( item_type == NULL )
	{
		libcerror_error_set(
		 error,
		 LIBCERROR_ERROR_DOMAIN_ARGUMENTS,
		 LIBCERROR_ARGUMENT_ERROR_INVALID_VALUE,
		 "%s: invalid item type.",
		 function );

		return( -1 );
	}
	*item_type = internal_item->type;

	return( 1 );
}

